import json
import logging

import pytest
from botocore.exceptions import ClientError
from fixtures.localstack import *  # noqa:F401
from opentaskpy.config.loader import ConfigLoader
from pytest_shell import fs

from opentaskpy.plugins.lookup.aws.ssm import run

PLUGIN_NAME = "ssm"

# Get the default logger and set it to DEBUG
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)


def test_ssm_plugin_missing_name():
    with pytest.raises(Exception) as ex:
        run()

    assert (
        ex.value.args[0]
        == f"Missing kwarg: 'name' while trying to run lookup plugin '{PLUGIN_NAME}'"
    )


def test_ssm_plugin_param_name_not_found(credentials):
    with pytest.raises(ClientError) as ex:
        run(name="does_not_exist")

    assert ex.value.response["Error"]["Code"] == "ParameterNotFound"


def test_ssm_plugin_standard_string(ssm_client):
    expected = "test1234"
    # Populate the SSM parameter store with a test value
    ssm_client.put_parameter(
        Name="my_test_param",
        Value=expected,
        Type="String",
        Overwrite=True,
    )

    result = run(name="my_test_param")

    assert result == expected


def test_ssm_plugin_secure_string(ssm_client):
    expected = "securetest1234"
    # Populate the SSM parameter store with a test value
    ssm_client.put_parameter(
        Name="my_secure_test_param",
        Value=expected,
        Type="SecureString",
        Overwrite=True,
    )

    result = run(name="my_secure_test_param")

    assert result == expected


def test_config_loader_using_ssm_plugin(ssm_client, tmpdir):
    json_obj = {
        "testLookup": "{{ lookup('aws.ssm', name='my_test_param') }}",
    }

    fs.create_files(
        [
            {
                f"{tmpdir}/variables.json.j2": {
                    "content": json.dumps(json_obj),
                }
            },
        ]
    )
    expected_result = "config_loader_test_1234"

    # Insert the param into paramstore
    ssm_client.put_parameter(
        Name="my_test_param",
        Value=expected_result,
        Type="String",
        Overwrite=True,
    )

    # Test that the global variables are loaded correctly
    config_loader = ConfigLoader(tmpdir)
    config_loader._load_global_variables()
    config_loader._resolve_templated_variables()

    assert config_loader.get_global_variables()["testLookup"] == expected_result



import glob
import os
import re

import boto3
import opentaskpy.logging
from opentaskpy.remotehandlers.remotehandler import (
    RemoteExecutionHandler,
    RemoteTransferHandler,
)

from .aws import set_aws_creds

MAX_OBJECTS_PER_QUERY = 100


class S3Transfer(RemoteTransferHandler):
    TASK_TYPE = "T"

    def __init__(self, spec):
        self.spec = spec

        self.logger = opentaskpy.logging.init_logging(
            __name__, os.environ.get("OTF_TASK_ID"), self.TASK_TYPE
        )

        set_aws_creds(self)

        # Use boto3 to setup the required object for the transfer
        kwargs = {
            "aws_access_key_id": self.aws_access_key_id,
            "aws_secret_access_key": self.aws_secret_access_key,
            "region_name": self.region_name,
        }
        # If there's an override for endpoint_url in the environment, then use that
        if os.environ.get("AWS_ENDPOINT_URL"):
            kwargs["endpoint_url"] = os.environ.get("AWS_ENDPOINT_URL")

        self.s3_client = boto3.client("s3", **kwargs)

    def handle_post_copy_action(self, files):
        # Determine the action to take
        # Delete the files
        if self.spec["postCopyAction"]["action"] == "delete":
            self.s3_client.delete_objects(
                Bucket=self.spec["bucket"],
                Delete={
                    "Objects": [{"Key": file} for file in files],
                    "Quiet": True,
                },
            )
        # Copy the files to the new location, and then delete the originals
        if (
            self.spec["postCopyAction"]["action"] == "move"
            or self.spec["postCopyAction"]["action"] == "rename"
        ):
            for file in files:
                # If this is a rename, then we need to determine the new key
                new_file = file
                if self.spec["postCopyAction"]["action"] == "rename":
                    new_file = f"{self.spec['postCopyAction']['destination']}{file.split('/')[-1]}"

                    # Use the pattern and sub values to rename the file correctly
                    new_file = re.sub(
                        self.spec["postCopyAction"]["pattern"],
                        self.spec["postCopyAction"]["sub"],
                        new_file,
                    )

                self.s3_client.copy_object(
                    Bucket=self.spec["bucket"],
                    CopySource={
                        "Bucket": self.spec["bucket"],
                        "Key": file,
                    },
                    Key=f"{self.spec['postCopyAction']['destination']}/{new_file}",
                )
                self.s3_client.delete_objects(
                    Bucket=self.spec["bucket"],
                    Delete={
                        "Objects": [{"Key": file}],
                        "Quiet": True,
                    },
                )
        return 0

    def list_files(self, directory=None, file_pattern=None):
        kwargs = {
            "Bucket": self.spec["bucket"],
            "MaxKeys": MAX_OBJECTS_PER_QUERY,
        }
        if directory:
            kwargs["Prefix"] = directory

        remote_files = dict()

        while True:
            response = self.s3_client.list_objects_v2(**kwargs)

            if response["KeyCount"]:
                for object in response["Contents"]:
                    key = object["Key"]
                    # Get the filename from the key
                    filename = key.split("/")[-1]  #
                    self.logger.debug(f"Found file: {filename}")
                    if file_pattern and not re.match(file_pattern, filename):
                        continue
                    else:
                        # Get the size and modified time
                        file_attr = self.s3_client.head_object(
                            Bucket=self.spec["bucket"], Key=key
                        )

                        remote_files[key] = {
                            "size": file_attr["ContentLength"],
                            "modified_time": file_attr["LastModified"].timestamp(),
                        }

                # This handles the pagination
                # if the NextContinuationToken doesnt exist, then we'll break out
                # of the loop
                try:
                    kwargs["ContinuationToken"] = response["NextContinuationToken"]
                except KeyError:
                    break
            else:
                return None

        return remote_files

    def move_files_to_final_location(self, files):
        raise NotImplementedError()

    # When S3 is the destination
    def pull_files(self, files, remote_spec):
        raise NotImplementedError()

    def push_files_from_worker(self, local_staging_directory):
        result = 0
        files = glob.glob(f"{local_staging_directory}/*")
        kwargs = {}
        if self.bucket_owner_full_control:
            kwargs["ACL"] = "bucket-owner-full-control"

        for file in files:
            # Strip the directory from the file
            file_name = file.split("/")[-1]
            self.logger.debug(
                f"Transferring file: {file} to s3://{self.spec['bucket']}/{file_name}"
            )
            try:
                self.s3_client.upload_file(
                    file,
                    self.spec["bucket"],
                    f"{self.spec['directory']}/{file_name}",
                    ExtraArgs=kwargs,
                )
            except Exception as e:
                self.logger.error(f"Failed to transfer file: {file}")
                self.logger.error(e)
                result = 1

        return result

    def pull_files_to_worker(self, files, local_staging_directory):
        result = 0
        for file in files:
            # Strip the directory from the file
            file_name = file.split("/")[-1]
            self.logger.debug(f"Transferring file: {file}")
            try:
                self.s3_client.download_file(
                    self.spec["bucket"],
                    file,
                    f"{local_staging_directory}/{file_name}",
                )
            except Exception as e:
                self.logger.error(f"Failed to transfer file: {file}")
                self.logger.error(e)
                result = 1

        return result

    def transfer_files(self, files, remote_spec, dest_remote_handler=None):
        # Check the remote handler, if it's another S3Transfer, then it's simple
        # to do an S3 copy via boto

        result = 0
        for file in files:
            # Strip the directory from the file
            file_name = file.split("/")[-1]
            self.logger.debug(f"Transferring file: {file}")
            try:
                self.s3_client.copy(
                    {
                        "Bucket": self.spec["bucket"],
                        "Key": file,
                    },
                    dest_remote_handler.spec["bucket"],
                    f"{dest_remote_handler.spec['directory']}/{file_name}",
                )
            except Exception as e:
                self.logger.error(f"Error transferring file: {file}")
                self.logger.error(e)
                result = 1

        return result

    def create_flag_files(self):
        result = 0

        object_key = self.spec["flags"]["fullPath"]
        self.logger.debug(f"Creating flag file: {object_key}")
        kwargs = {
            "Bucket": self.spec["bucket"],
            "Key": object_key,
        }
        if self.bucket_owner_full_control:
            kwargs["ACL"] = "bucket-owner-full-control"

        try:
            self.s3_client.put_object(**kwargs)
        except Exception as e:
            self.logger.error(f"Error creating flag file: {object_key}")
            self.logger.error(e)
            result = 1

        return result

    def tidy(self):
        self.s3_client.close()


class S3Execution(RemoteExecutionHandler):
    TASK_TYPE = "E"

    def tidy(self):
        self.s3_client.close()

    def __init__(self, spec):
        self.spec = spec

        self.logger = opentaskpy.logging.init_logging(
            __name__, os.environ.get("OTF_TASK_ID"), self.TASK_TYPE
        )

        set_aws_creds(self)

        # Use boto3 to setup the required object
        kwargs = {
            "aws_access_key_id": self.aws_access_key_id,
            "aws_secret_access_key": self.aws_secret_access_key,
            "region_name": self.region_name,
        }
        # If there's an override for endpoint_url in the environment, then use that
        if os.environ.get("AWS_ENDPOINT_URL"):
            kwargs["endpoint_url"] = os.environ.get("AWS_ENDPOINT_URL")

        self.s3_client = boto3.client("s3", **kwargs)

    # This cannot be long running, so kill doesnt really need to do anything
    def kill(self):
        pass

    def execute(self):
        result = True

        object_key = self.spec["key"]
        bucket = self.spec["bucket"]

        kwargs = {
            "Bucket": bucket,
            "Key": object_key,
        }
        if self.bucket_owner_full_control:
            kwargs["ACL"] = "bucket-owner-full-control"

        # Use the boto client to write a blank file with this name to the bucket
        try:
            self.s3_client.put_object(**kwargs)
        except Exception as e:
            self.logger.error(f"Failed to create flag file: {object_key}")
            self.logger.error(e)
            result = False

        return result



"""
SSM Param Store lookup plugin

Uses boto3 to pull out a parameter from AWS Parameter Store
This uses AWS authentication, either from the IAM role of the host,
by using environment variables, or variables in variables.json file

"""
import os

import boto3
import opentaskpy.logging
from botocore.exceptions import ClientError

logger = opentaskpy.logging.init_logging(__name__)

plugin_name = "ssm"


def run(**kwargs):
    # Expect a kwarg named url, and value
    expected_kwargs = ["name"]
    for kwarg in expected_kwargs:
        if kwarg not in kwargs:
            raise Exception(
                f"Missing kwarg: '{kwarg}' while trying to run lookup plugin '{plugin_name}'"
            )

    globals = kwargs["globals"] if "globals" in kwargs else None

    aws_access_key_id = (
        globals["AWS_ACCESS_KEY_ID"]
        if globals and "AWS_ACCESS_KEY_ID" in globals
        else os.environ.get("AWS_ACCESS_KEY_ID")
    )
    aws_secret_access_key = (
        globals["AWS_SECRET_ACCESS_KEY"]
        if globals and "AWS_SECRET_ACCESS_KEY" in globals
        else os.environ.get("AWS_SECRET_ACCESS_KEY")
    )

    region_name = (
        globals["AWS_REGION"]
        if globals and "AWS_REGION" in globals
        else os.environ.get("AWS_REGION")
    )
    boto3_kwargs = {
        "aws_access_key_id": aws_access_key_id,
        "aws_secret_access_key": aws_secret_access_key,
        "region_name": region_name,
    }
    # If there's an override for endpoint_url in the environment, then use that
    if os.environ.get("AWS_ENDPOINT_URL"):
        boto3_kwargs["endpoint_url"] = os.environ.get("AWS_ENDPOINT_URL")

    result = None
    try:
        ssm = boto3.client("ssm", **boto3_kwargs)
        response = ssm.get_parameter(Name=kwargs["name"], WithDecryption=True)
        result = response["Parameter"]["Value"]

        logger.log(12, f"Read '{result}' from param {kwargs['name']}")

    except ClientError as e:
        if e.response["Error"]["Code"] == "ParameterNotFound":
            logger.error(f"Parameter not found: {kwargs['name']}: {e}")
            raise e
        else:
            raise e
    except Exception as e:
        logger.error(f"Failed to read from SSM parameter: {kwargs['name']}: {e}")

    return result



import io
import json
import logging
import os
import re
import subprocess
import time
import zipfile

import botocore.exceptions
import opentaskpy.logging
import pytest
from fixtures.localstack import *  # noqa:F401
from opentaskpy.config.loader import ConfigLoader
from opentaskpy.taskhandlers import batch, execution
from pytest_shell import fs

os.environ["OTF_NO_LOG"] = "0"
os.environ["OTF_LOG_LEVEL"] = "DEBUG"

BUCKET_NAME = "otf-addons-aws-lambda-execution-test"
BUCKET_NAME_1 = "otf-addons-aws-lambda-execution-test-1"
BUCKET_NAME_2 = "otf-addons-aws-lambda-execution-test-2"
logger = opentaskpy.logging.init_logging(__name__)

logger.setLevel(logging.DEBUG)

root_dir_ = get_root_dir()

lambda_execution_task_definition = {
    "type": "execution",
    "functionArn": None,
    "payload": {"bucket_name": BUCKET_NAME, "file_name": "function_test.txt"},
    "protocol": {
        "name": "opentaskpy.addons.aws.remotehandlers.lambda.LambdaExecution",
    },
}

lambda_batch_task_definition = {
    "type": "batch",
    "tasks": [
        {
            "order_id": 1,
            "task_id": "lambda_execution_task_definition",
            "timeout": 2,
        }
    ],
}


def create_lambda_function(lambda_client, lambda_handler, payload, invoke=True):
    # Read in lambda_handler
    lambda_code = None
    with open(os.path.join(root_dir_, "../tests", lambda_handler), "rb") as f:
        lambda_code = f.read()

    zip_buffer = io.BytesIO()
    # Write a zip file into the buffer
    with zipfile.ZipFile(zip_buffer, "w", zipfile.ZIP_DEFLATED) as zf:
        info = zipfile.ZipInfo(lambda_handler)
        info.compress_type = zipfile.ZIP_DEFLATED
        info.external_attr = 0o777 << 16  # give full access
        zf.writestr(info, lambda_code)
    zip_buffer.seek(0)

    # Delete existing function
    try:
        lambda_client.delete_function(FunctionName="my-function")
    except Exception:
        pass

    # Create a lambda function that creates a file on our test S3 bucket
    lambda_response = lambda_client.create_function(
        FunctionName="my-function",
        Runtime="python3.9",
        Code={
            "ZipFile": zip_buffer.read(),
        },
        Handler=f"{re.sub('.py', '', lambda_handler)}.lambda_handler",
        Role="arn:aws:iam::123456789012:role/lambda-role",
        Timeout=300,
        MemorySize=128,
    )
    function_arn = lambda_response["FunctionArn"]

    # Manually call the function to check it's actually working
    if invoke:
        lambda_client.invoke(
            FunctionName=function_arn,
            InvocationType="RequestResponse",
            Payload=json.dumps(payload),
        )
        # Make sure the response is a 200
        assert lambda_response["ResponseMetadata"]["HTTPStatusCode"] == 200

    return function_arn


@pytest.fixture(scope="session")
def setup_bucket(credentials):
    # This all relies on docker container for the AWS stack being set up and running
    # The AWS CLI should also be installed

    buckets = [BUCKET_NAME, BUCKET_NAME_1, BUCKET_NAME_2]
    # Delete existing buckets and recreate
    for bucket in buckets:
        subprocess.run(["awslocal", "s3", "rb", f"s3://{bucket}", "--force"])
        subprocess.run(["awslocal", "s3", "mb", f"s3://{bucket}"])


def test_remote_handler():
    execution_obj = execution.Execution(
        "call-lambda-function", lambda_execution_task_definition
    )

    execution_obj._set_remote_handlers()

    # Validate some things were set as expected
    assert execution_obj.remote_handlers[0].__class__.__name__ == "LambdaExecution"


def test_run_lambda_function(setup_bucket, lambda_client, s3_client):
    function_arn = create_lambda_function(
        lambda_client,
        "lambda_write_to_s3.py",
        {
            "bucket_name": BUCKET_NAME,
            "file_name": "function_test.txt",
        },
    )

    # Check that the file has been created in the bucket
    s3_response = s3_client.head_object(
        Bucket=BUCKET_NAME,
        Key="function_test.txt",
    )
    assert s3_response["ResponseMetadata"]["HTTPStatusCode"] == 200

    # Delete the object from the bucket
    s3_client.delete_object(
        Bucket=BUCKET_NAME,
        Key="function_test.txt",
    )

    with pytest.raises(botocore.exceptions.ClientError) as ex:
        s3_client.head_object(
            Bucket=BUCKET_NAME,
            Key="function_test.txt",
        )
    # Check the exception is a 404
    assert ex.value.response["Error"]["Code"] == "404"

    # Update the task definition with the ARN of the function we just created
    lambda_execution_task_definition_copy = lambda_execution_task_definition.copy()
    lambda_execution_task_definition_copy["functionArn"] = function_arn
    lambda_execution_task_definition_copy["payload"]["bucket"] = BUCKET_NAME

    # Call the execution and check whether the lambda function ran successfully
    execution_obj = execution.Execution(
        "call-lambda-function", lambda_execution_task_definition_copy
    )

    logging.getLogger("boto3").setLevel(logging.DEBUG)
    logging.getLogger("botocore").setLevel(logging.DEBUG)

    assert execution_obj.run()

    # Wait a second for the lambda to finish
    time.sleep(2)

    # Check that the lambda function has run, and created the file
    s3_response = s3_client.head_object(
        Bucket=BUCKET_NAME,
        Key="function_test.txt",
    )
    assert s3_response["ResponseMetadata"]["HTTPStatusCode"] == 200

    # Invoke it again, but using the RequestResponse invocation type
    # This should return a 200 response

    # clone lambda_execution_task_definition
    lambda_execution_task_definition_copy = lambda_execution_task_definition.copy()
    lambda_execution_task_definition_copy["functionArn"] = function_arn
    lambda_execution_task_definition_copy["payload"]["bucket"] = BUCKET_NAME
    lambda_execution_task_definition_copy["invocationType"] = "RequestResponse"
    execution_obj = execution.Execution(
        "call-lambda-function", lambda_execution_task_definition_copy
    )
    assert execution_obj.run()


def test_run_lambda_function_with_invalid_payload(lambda_client):
    # Create the lambda function
    function_arn = create_lambda_function(
        lambda_client,
        "lambda_write_to_s3.py",
        {
            "bucket_name": BUCKET_NAME_1,
            "file_name": "function_test.txt",
        },
    )

    lambda_execution_task_definition_invalid_payload = (
        lambda_execution_task_definition.copy()
    )
    # Update the task definition with the ARN of the function we just created
    lambda_execution_task_definition_invalid_payload["functionArn"] = function_arn
    # Remove the payload from the definition
    lambda_execution_task_definition_invalid_payload.pop("payload")
    lambda_execution_task_definition_invalid_payload[
        "invocationType"
    ] = "RequestResponse"

    # Call the execution and check whether the lambda function ran successfully
    execution_obj = execution.Execution(
        "call-lambda-function", lambda_execution_task_definition_invalid_payload
    )

    assert not execution_obj.run()

    # Try the same, but using an async call to the function
    lambda_execution_task_definition_invalid_payload["invocationType"] = "Event"
    execution_obj = execution.Execution(
        "call-lambda-function", lambda_execution_task_definition_invalid_payload
    )

    # This is expected to work, because there's no validation of the payload when invoked this way
    assert execution_obj.run()


def test_lambda_execution_batch_timeout(tmpdir, lambda_client):
    # Create the function, but dont invoke it, as it runs too long
    function_arn = create_lambda_function(
        lambda_client, "lambda_sleep_60.py", None, invoke=False
    )

    # set the arn in lambda_execution_task_definition
    lambda_execution_task_definition["functionArn"] = function_arn

    # We need to write the lambda execution defintion to a file for the config_loader to read from
    with open(f"{tmpdir}/lambda_execution_task_definition.json", "w") as f:
        json.dump(lambda_execution_task_definition, f)

    # Write a dummy variables file
    fs.create_files([{f"{tmpdir}/variables.json": {"content": "{}"}}])

    config_loader = ConfigLoader(tmpdir)

    # Enable logging
    del os.environ["OTF_NO_LOG"]

    batch_obj = batch.Batch("timeout", lambda_batch_task_definition, config_loader)
    assert batch_obj.run()



"""Setup function for otf-addons-aws package."""

from setuptools import find_namespace_packages, setup

setup(
    name="otf-addons-aws",
    packages=find_namespace_packages(where="src"),
    package_dir={"": "src"},
)



import base64
import json
import os

import boto3
import botocore.exceptions
import opentaskpy.logging
from opentaskpy.remotehandlers.remotehandler import RemoteExecutionHandler

from .aws import set_aws_creds


class LambdaExecution(RemoteExecutionHandler):
    TASK_TYPE = "E"

    def tidy(self):
        self.lambda_client.close()

    def __init__(self, spec):
        self.spec = spec

        # Ensure that function_arn is defined in the spec
        if "functionArn" not in self.spec:
            raise Exception("functionArn not defined in spec")

        self.logger = opentaskpy.logging.init_logging(
            __name__, os.environ.get("OTF_TASK_ID"), self.TASK_TYPE
        )

        set_aws_creds(self)

        # Use boto3 to setup the required object
        kwargs = {
            "aws_access_key_id": self.aws_access_key_id,
            "aws_secret_access_key": self.aws_secret_access_key,
            "region_name": self.region_name,
        }
        # If there's an override for endpoint_url in the environment, then use that
        if os.environ.get("AWS_ENDPOINT_URL"):
            kwargs["endpoint_url"] = os.environ.get("AWS_ENDPOINT_URL")

        # Set the client to only try once. This prevents the lambda function being called more than once
        # and issues when working with batch timeouts if a lambda function takes longer than the batch timeout
        # It still means that the timeout of a lambda function is always at least 60 seconds due to the way boto3's HTTP timeout works
        kwargs["config"] = botocore.client.Config(retries={"max_attempts": 0})

        self.lambda_client = boto3.client("lambda", **kwargs)

    def kill(self):
        # We need to kill the running execute function
        self.tidy()
        self.logger.info("Killing thread")

    # Triggers the lambda function. The synchronous invocation type will cause this thread
    # to block. This can be an issue with batch jobs and their timeout.
    # Boto3 will timeout after 60 seconds regardless, so Lambda functions that are called should not be long running. Their own timeout should be less than the
    # batch timeout if one is used.

    # An async call will not check the status of the lambda function, only if there are errors with invoking it.

    def execute(self):
        result = True

        function_arn = self.spec["functionArn"]
        invocation_type = (
            self.spec["invocationType"] if "invocationType" in self.spec else "Event"
        )
        payload = None
        if "payload" in self.spec:
            payload = self.spec["payload"]

        try:
            invoke_response = self.lambda_client.invoke(
                FunctionName=function_arn,
                InvocationType=invocation_type,
                Payload=json.dumps(payload),
            )

            if (
                invoke_response["StatusCode"] != 200
                and invoke_response["StatusCode"] != 202
            ):
                self.logger.error(f"Failed to run lambda function: {function_arn}")
                return False

            if "FunctionError" in invoke_response:
                self.logger.error(
                    f"Lambda function returned an error: {function_arn} - AWS Exception message: {invoke_response['FunctionError']}"
                )
                return False

            # Log the response if there is one
            if "LogResult" in invoke_response:
                # base64 decode the result
                log_result = base64.b64decode(invoke_response["LogResult"])

                self.logger.debug(f"Lambda function response: {log_result}")

        except botocore.exceptions.ClientError as e:
            self.logger.error(f"Failed to run lambda function: {function_arn}")
            self.logger.error(e)
            result = False

        return result
